subroutine map_parameters(task,map,hiuv,freq,uvmax,uvmin,error,print)
  use gkernel_interfaces
  use imager_interfaces, only : map_message, telescope_beam
  use clean_def
  use clean_arrays
  use clean_types
  use clean_default
  use gbl_message
  !----------------------------------------------------------------------
  !   @ private-mandatory
  !*
  ! IMAGER -- Utility routine 
  !
  !   Prepare the MAP parameters for UV_MAP, MX or UV_RESTORE  
  ! Reads these from SIC variables if present, or guess them from UV data
  !!
  !-----------------------------------------------------------------------
  character(len=*), intent(in)    :: task   !! Input task name (UV_MAP, UV_SELF, UV_RESTORE, or MX)
  type (uvmap_par), intent(inout) :: map    !! Map parameters
  type (gildas), intent(in) :: hiuv         !! UV header
  real(8), intent(inout) :: freq            !! Observing frequency
  real(4), intent(in) :: uvmax, uvmin       !! Min & Max UV in 2 pi Wavelengths
  logical, intent(inout) :: error           !! Logical error flag
  logical, optional :: print                !! Printout option
  !
  ! Constants
  real(8), parameter :: pi=3.14159265358979323846d0
  real(8), parameter :: sec_in_rad = pi/3600d0/180d0
  real(8), parameter :: f_to_k = 2.d0*pi/299792458.d-6
  real(4), parameter :: pixel_per_beam=5.0
  !
  ! Local ---
  real(4) :: map_range(2)  ! Range of offsets
  character(len=96) :: chain
  character(len=12) :: prefix
  real :: uvmi, uvma, xp, xm, hppb, bsize
  integer :: i, np
  type (uvmap_par), save :: last_map
  logical :: do_print, more
  integer :: osize(2)
  !
  character(len=32) :: rname
  integer(kind=8) :: ram_uv, ram_map, ram_need
  integer :: map_size(2)
  real :: map_field(2), map_xycell(2)
  !
  ! Code ----
  do_print = present(print)
  !
  ! Remember input paramaters
  map_size = map%size
  map_field = map%field
  map_xycell = map%xycell
  !
  if (task.eq.'UV_SETUP') then
    rname = 'UV_STAT'
  else
    rname = task
  endif
  !
  freq = gdf_uv_frequency(hiuv)
  if (do_print) then                                            
    write(chain,'(A,F12.6,A)') "Observed rest frequency       ", &
      & freq*1d-3," GHz"
    call map_message(seve%i,rname,chain)
  endif
  !
  uvma = uvmax/(freq*f_to_k)
  uvmi = uvmin/(freq*f_to_k)
  if (do_print) then
    write(chain,100) hiuv%gil%nvisi,hiuv%gil%nchan
    call map_message(seve%i,rname,chain)
    write(chain,101) uvmi,uvma,' meters'
    call map_message(seve%i,rname,chain)
    write(chain,101) uvmin*1e-3/2.0/pi,uvmax*1e-3/2.0/pi,' kiloWavelength'
    call map_message(seve%i,rname,chain)
  endif
  !
  !
  ! Get beam size (in radians)
  bsize = telescope_beam(rname,hiuv)
  !
  if (bsize.eq.0) then
    !
    ! Default to IRAM Plateau de Bure case
    hppb = 63.0*(12.0/15.0)*(100.e3/freq) ! [arcsec] Primary beam size
    ! 1.22 Lambda/D for NOEMA 15 m antennas
    write(chain,'(A,F12.1,A)') "Half power primary beam defaulted to ", &
      & hppb," arcsec (15-m diameter antennas)"
    call map_message(seve%w,rname,chain)                         
    bsize = hppb*sec_in_rad
  else 
    if (hiuv%gil%nteles.gt.0) then
      recmap_uvcell = hiuv%gil%teles(1)%diam/2.0       
    else
      ! Derive first order UV cell size
      recmap_uvcell = 1.22*3E8/(freq*1E6) / 2. / bsize
    endif
    if (do_print .or. map%size(1).eq.0) then                                           
      write(chain,'(A,F12.1,A)') "Half power primary beam       ", &
        & bsize/sec_in_rad," arcsec"
      call map_message(seve%i,rname,chain) 
    endif
    !
  endif
  hppb = bsize/sec_in_rad 
  !
  ! Only 2 out of the 3 MAP_FIELD, MAP_SIZE and MAP_CELL
  !   can be non-zero. This should be tested in some way
  !
  if (map_xycell(1)*map_size(1)*map_field(1).ne.0) then
    if (abs(map_field(1)-map_size(1)*map_xycell(1)).gt.0.01*map_field(1)) then
      call map_message(seve%w,rname,'User specified MAP_FIELD, MAP_SIZE and MAP_CELL do not match')
    else
      call map_message(seve%w,rname,'MAP_FIELD, MAP_SIZE and MAP_CELL specified by user')    
    endif
  endif
  !
  ! Now get MAP_CELL and MAP_SIZE if Zero.
  ! Default Map_Cell, in ", rounded to 0.02 or 0.001"
  recmap_xycell = 0.01*nint(360.0*3600.0*100.0/uvmax/pixel_per_beam)
  if (recmap_xycell(1).le.0.02) then
    recmap_xycell = 0.001*nint(360.0*3600.0*1000.0/uvmax/pixel_per_beam)
  endif
  if (map_xycell(1).eq.0) then
    if (map_size(1)*map_field(1).eq.0) then
      map_xycell = recmap_xycell
    else
      map_xycell(1:2) = map_field(1)/map_size(1)
    endif
  endif
  !
  ! Get a wide enough field of view (in arcsec)
  if (map%nfields.ne.0) then
    ! The factor GUARD_SIZE on the primary beam size is to go down 
    ! low enough by default
    recmap_field = guard_size*hppb
    ! Enlarge the field of view by the Offset range
    np = abs(map%nfields)
    if (allocated(hmosaic)) then
      xm = minval(hmosaic(:)%opoint(1))
      xp = maxval(hmosaic(:)%opoint(1))
      map_range(1) = xp-xm
      xm = minval(hmosaic(:)%opoint(2))
      xp = maxval(hmosaic(:)%opoint(2))
      map_range(2) = xp-xm
    else
      xm = minval(map%offxy(1,1:np))
      xp = maxval(map%offxy(1,1:np))
      map_range(1) = xp-xm
      xm = minval(map%offxy(2,1:np))
      xp = maxval(map%offxy(2,1:np))
      map_range(2) = xp-xm
    endif
    map_range = map_range/sec_in_rad
    !
    recmap_field = recmap_field + map_range
  else
    ! The factor 2 on the primary beam size is to go down to 6.25 % level by default
    recmap_field = 2*hppb
    ! Map Field = 1.13*Lambda/Dmin - but round to 1.3
    if (uvmin.ne.0) then
      ! The Min is to avoid going to very large images when uvmin is close to 0
      ! e.g. when short or Zero spacings have been added
      recmap_field = min(recmap_field,2.6*360*3600/uvmin) ! In arcsec
    endif
  endif
  !
  ! Round recmap_field to nearest 0.1 arcsec
  recmap_field = 0.1d0*nint(10.d0*recmap_field)
  !
  ! Default MAP_SIZE
  recmap_size = recmap_field/recmap_xycell
  call gi4_round_forfft(recmap_size(1),osize(1),error,map_rounding,map_power)
  call gi4_round_forfft(recmap_size(2),osize(2),error,map_rounding,map_power)
  recmap_size = osize
  !
  ! Get MAP_SIZE if non Zero to define the actual Field of view
  if (map_size(1).ne.0) then
    if (map_size(2).eq.0) map_size(2) = map_size(1)
    if (map_field(1).eq.0) map_field = map_size * map_xycell
  endif
  if (map_field(1).ne.0) then
    if (map_field(2).eq.0) map_field(2) = map_field(1)
  else
    !
    map_field = recmap_field
  endif
  !
  if (map_size(1).eq.0) then
    ! At this stage, map_field and map_xycell are non zero. 
    map_size = map_field/map_xycell
    call gi4_round_forfft(map_size(1),osize(1),error,map_rounding,map_power)
    call gi4_round_forfft(map_size(2),osize(2),error,map_rounding,map_power)
    map_size = osize
    !
  else if ( (mod(map%size(1),2).ne.0) .or. (mod(map%size(2),2).ne.0) ) then
    call map_message(seve%e,rname,'Number of pixels (MAP_SIZE) must be even')
    error = .true.
  endif
  ! Round MAP%FIELD now
  map_field = 0.1*nint(10.0*map_size*map_xycell)
  !
  ! Get the results 
  !   Do NOT adjust the Field while keeping the Cell size 
  !     map%field = map%size * map%xycell
  ! it is useless or incorrect...
  !
  if (map%xycell(1).eq.0) map%xycell = map_xycell   ! Cell must be defined at this stage
  if (map%field(1).ne.0) then    
    ! Field imposes the size
    map%size = map%field/map%xycell
    call gi4_round_forfft(map%size(1),osize(1),error,map_rounding,map_power)
    call gi4_round_forfft(map%size(2),osize(2),error,map_rounding,map_power)
    map%size = osize
  else
    ! Size may be defined
    if (map%size(1).eq.0) then 
      map%field = map_field
      map%size = map_size
    else
      map%field = map%size * map%xycell
    endif
  endif
  !
  more = .false.
  if (task.eq.'UV_STAT') then
    prefix = 'Recommended '
    map_size = recmap_size
    map_field = recmap_field
    map_xycell = recmap_xycell
    more = .true.
  else if (task.eq.'UV_SETUP') then
    prefix = 'Adopted '
  else
    if (error) return
    prefix = 'Current '
  endif
  !
  write(chain,'(A,A,F10.1,a,F10.1,A)') prefix,'Field of view ' &
    & ,map_field(1),' x ',map_field(2),' "'
  if (more) then
    if (any(map%field.ne.recmap_field)) &
    write(chain(53:),'(A,F10.1,a,F10.1,A)') 'Current ' &
    & ,map%field(1),' x ',map%field(2),' "'
  endif
  call map_message(seve%i,rname,chain)
  !
  write(chain,'(A,A,F10.1,A,F10.1,A)') prefix,'Image size    ', &
    & map_size(1)*map_xycell(1),' x ',map_size(2)*map_xycell(2),' " '
  if (more) then
    if (any(map%size*map%xycell.ne.map_size*map_xycell)) &
    write(chain(53:),'(A,F10.1,a,F10.1,A)') 'Current ' &
    & ,map%size(1)*map%xycell(2),' x ',map%size(2)*map%xycell(2),' "'
  endif
  call map_message(seve%i,rname,chain)
  !
  write(chain,'(A,A,I10,A,I10,A)') prefix,'Map size      ', &
    & map_size(1),' x ',map_size(2)
  if (more) then
    if (any(map%size.ne.map_size)) &
    write(chain(53:),'(A,I10,a,I10,A)') 'Current ' &
    & ,map%size(1),' x ',map%size(2)
  endif
  call map_message(seve%i,rname,chain)
  !
  write(chain,'(A,A,F10.3,A,F10.3,A)') prefix,'Pixel size    ', &
    & map_xycell(1),' x ',map_xycell(2),' "'
  if (more) then
    if (any(map%xycell.ne.map_xycell)) &
    write(chain(53:),'(A,F10.3,a,F10.3,A)') 'Current ' &
    & ,map%xycell(1),' x ',map%xycell(2),' "'
  endif
  call map_message(seve%i,rname,chain)
  !
  ! Evaluate RAM pressure
  ram_uv = (huv%gil%dim(1)*huv%gil%dim(2))/(512*512)   ! Size of RAM requested
  ram_map = (2*map%size(1)*map%size(2))/(512*512)*huv%gil%nchan
  ram_need = ram_map+ram_uv
  !!Print *,'UV size ',ram_uv, huv%gil%dim(1:2), huv%gil%dim(1)*huv%gil%dim(2)
  !!Print *,'Map size ',ram_map, huv%gil%nchan, map%size, map%size(1)*map%size(2)
  if (ram_need.gt.sys_ramsize) then
      write(chain,'(A,F8.1,A,F8.1,A,F8.1,A)') 'Data size (UV ',1d-3*ram_uv,& 
      & 'GB + Map ',1D-3*ram_map, &
      & 'GB) exceeds available RAM (',1d-3*sys_ramsize,' GB)'
    call map_message(seve%e,rname,chain,1)
    error = .true.
    return
  else if (ram_need.gt.sys_ramsize/3) then
    write(chain,'(A,F8.1,A,F8.1,A,F8.1,A)') 'Data size (UV ',1d-3*ram_uv,& 
      & 'GB + Map ',1D-3*ram_map, &
      & 'GB) exceeds 1/3rd of available RAM (',1d-3*sys_ramsize,' GB)'
    call map_message(seve%w,rname,chain,3)
  else if (rname.eq.'UV_MAP') then
    write(chain,'(A,F8.1,A,F8.1,A,F8.1,A)') 'Data size (UV ',1d-3*ram_uv,& 
      & 'GB + Map ',1D-3*ram_map, &
      & 'GB)  for a RAM of (',1d-3*sys_ramsize,' GB)'
    call map_message(seve%i,rname,chain,3)
  endif
  !  
  if (rname.ne.'UV_STAT') then
!  else
    ! This part is only for UV_MAP / UV_MOSAIC, not for UV_STAT
    !
    ! Check whether Weights have to be recomputed
    if (map%uniform(1).ne.last_map%uniform(1)) do_weig = .true.
    if (map%uniform(2).ne.last_map%uniform(2)) do_weig = .true.
    if (map%ctype.ne.last_map%ctype) do_weig = .true.
    do i=1,4
      if (map%taper(i).ne.last_map%taper(i)) do_weig = .true.
    enddo
    last_map = map
  endif
  !
100 format('Found ',i12,' Visibilities, ',i4,' channels')
101 format('Baselines ',f9.1,' - ',f9.1,a)
end subroutine map_parameters
!
!<FF>
!
subroutine uv_stat_comm(line,error)
  use gkernel_interfaces
  use imager_interfaces, except_this=>uv_stat_comm
  use clean_def
  use clean_default
  use clean_arrays
  use gbl_message
  !------------------------------------------------------------------------
  ! @ private
  !*
  ! IMAGER  --  Support for command UV_STAT [/PLOT]
  !
  !       Analyse a UV data set to define approximate beam size,
  !       field of view, expected "best beam", etc...
  !
  ! Optionally plot the corresponding dirty beams.
  !!
  !------------------------------------------------------------------------
  character(len=*), intent(inout) :: line  !! Command line
  logical, intent(out) :: error            !! Logical error flag
  !
  ! Constants
  real(kind=8), parameter :: pi=3.14159265358979323846d0
  real(kind=8), parameter :: f_to_k = 2.d0*pi/299792458.d-6
  integer, parameter :: o_plot=1
  integer, parameter :: mt=100
  integer, parameter :: mmode=13
  character(len=8) :: smode(mmode)
  data smode /'ADVISE','ALL','BASELINE','BEAM','BRIGGS','CELL','DEFAULT','HEADER', &
    & 'RESET','ROBUST','SETUP','TAPER','WEIGHT'/
  character(len=7) :: rname = 'UV_STAT'
  !
  ! Local ---
  real(4), save :: bmax, bmin
  logical, save :: first = .true.
  real, allocatable, save :: result(:,:)
  logical, save :: old_exist=.false.
  integer, save :: muni=0
  type (uvmap_par), save :: amap    ! To get the New variable names
  type (uvmap_par), save :: oldmap  ! To save the previous MAP_...
!  type(gildas), save :: hb          ! Beam Header
!  real, allocatable, target, save :: beams(:,:,:) ! Should it be saved ?
  !
  integer datelist(mt)
  character(len=128) :: fich
  logical :: do_plot
  integer :: jcol
  !
  integer ndates, nuni
  character(len=8) mode, argum
  real, allocatable :: fft(:)
  real uv_taper(3),map_cell(2),uniform(2),uvmax,uvmin,uvm
  real start,step,taper,robust
  integer map_size(2),wcol,ctype,n,mcol(2),ier
  integer(kind=index_length) :: dim(4)
  logical sorted, shift
  real(kind=8) :: freq, new(3)
  logical :: print
  character(len=message_length) :: chain
  !
  ! Code ----
  if (sic_present(0,1)) then
    call sic_ke(line,0,1,argum,n,.false.,error)
    if (error) return
    call sic_ambigs (rname,argum,mode,n,smode,mmode,error)
    if (error) return
  else
    mode = 'ALL'
  endif
  do_plot = sic_present(o_plot,0)
  !
  freq = gdf_uv_frequency(huv)
  !
  select case (mode)
  case ('DEFAULT')
    default_map%size = 0
    default_map%xycell = 0.
    default_map%field = 0.
    default_map%uvcell = 0 
    return
  case ('CELL','TAPER','WEIGHT','BRIGGS','ROBUST')
    if ((mode.eq.'CELL').and.(default_map%uniform(2).eq.0)) then
      call map_message(seve%e,rname,'Null MAP_ROBUST factor, cannot use MAP_UVCELL')        
      error = .true.
      return
    endif
    wcol = 0
    mcol = 0
    !
    start = 0.0
    call sic_r4(line,0,2,start,.false.,error)
    start = abs(start)
    step = 0.0
    call sic_r4(line,0,3,step,.false.,error)
    step = abs(step)
    !
    ctype = 1
    !
    ! Must do like UV_MAP now
    robust = default_map%uniform(2) ! Save state
    if ((mode.eq.'WEIGHT').or.(mode.eq.'ROBUST')) then
      default_map%uniform(2) = 1.0  ! Non zero to have the "Robust" message
    else if (mode.eq.'BRIGGS') then
      default_map%uniform(2) = -1.0 ! Non zero to have the "Robust" message
    endif
    call map_prepare(rname,huv,amap,error)
    default_map%uniform(2) = robust ! Restore state
    if (error) return
    !
    uniform(:) = amap%uniform(:)
    uv_taper(:) = amap%taper(1:3) 
    call sic_get_inte('WCOL',wcol,error)
    if (error) return
    !
    ! First sort the input UV Table, leaving UV Table in UV_*
    shift = .false.
    call uv_sort (huv,duv,error,sorted,shift,new,uvmax,uvmin,.false.)
    if (error) return
  case ('BEAM')
    wcol = 0
    mcol = 0
    call map_prepare(rname,huv,amap,error)
    uniform(:) = amap%uniform(:)
    uv_taper(:) = amap%taper(1:3) 
    call sic_get_inte('WCOL',wcol,error)
    if (error) return
    !
    ! First sort the input UV Table, leaving UV Table in UV_*
    shift = .false.
    call uv_sort (huv,duv,error,sorted,shift,new,uvmax,uvmin,.false.)
    if (error) return
  case ('ADVISE','ALL','HEADER','SETUP','BASELINE')
    ! For HEADERS, do the minimum work for speed...
    ! The UV table is available in HUV%
    if (huv%loca%size.eq.0) then
      call map_message(seve%e,rname,'No UV data loaded')
      error = .true.
      return
    endif
    call uvgmax (huv,duv,uvmax,uvmin)
    ! Now transform UVMAX in kiloWavelength (including 2 pi factor)
    uvmax = uvmax*freq*f_to_k
    uvmin = uvmin*freq*f_to_k
    error = .false.
  end select
  !
  ! New code: For the time being, this is somewhat inconsistent
  ! with the use of uv_sort which uses a slightly different frequency.
  ! It does not matter, since this routine provides hints, not
  ! exact values.
  !
  bmax = uvmax/(freq*f_to_k)
  bmin = uvmin/(freq*f_to_k)
  if (first) then
!    call sic_defstructure('UVSTAT',.true.,error)
!    call sic_def_real('UVSTAT%BMAX',bmax,0,dim,.false.,error)
!    call sic_def_real('UVSTAT%BMIN',bmin,0,dim,.false.,error)
    call sic_def_real('UV_BMAX',bmax,0,dim,.false.,error)
    call sic_def_real('UV_BMIN',bmin,0,dim,.false.,error)
    !
    call sic_defstructure('DEFMAP',.true.,error)
    call sic_def_inte('DEFMAP%SIZE',themap%size,1,2,.true.,error)
    call sic_def_real('DEFMAP%CELL',themap%xycell,1,2,.true.,error)
    call sic_def_real('DEFMAP%FIELD',themap%field,1,2,.true.,error)
    !
    first = .false.
  endif
  !
  select case (mode)
  case ('BASELINE') 
    do_plot = .true.
    call uv_get_cols(rname,huv,duv,huv%gil%nchan,mcol,wcol, jcol, freq, error)
    if (error) return
    jcol = 7+3*jcol
    call uv_stat_base(rname,huv,duv,bmin,bmax,jcol,do_plot,error)
  case ('BEAM') 
    call map_beamsize(amap,huv,duv,bmin,bmax,do_plot,error)
  case ('HEADER')
    call gdf_print_header(huv)
    call uv_listheader(huv,duv,mt,datelist,ndates,freq)
    call uv_printoffset('HEADER',huv,duv,themap)
  case ('ADVISE')
    call map_copy_par(default_map,themap)
    call map_parameters(rname,themap,huv,freq,uvmax,uvmin,error,print)
    error = .false.
  case ('ALL')
    call gdf_print_header(huv)
    call uv_listheader(huv,duv,mt,datelist,ndates,freq)
    call uv_printoffset('HEADER',huv,duv,themap)
    call map_copy_par(default_map,themap)
    !!Print *,'THEMAP ',themap%size,themap%field,themap%xycell
    call map_parameters(rname,themap,huv,freq,uvmax,uvmin,error,print)
  case ('RESET')
    if (old_exist) then
      call map_copy_par(oldmap,default_map) 
      old_exist = .false.
    endif
  case ('SETUP')
    call map_copy_par(default_map,themap)
    call map_parameters('UV_SETUP',themap,huv,freq,uvmax,uvmin,error,print)
    if (error) return
    call map_copy_par(default_map,oldmap)
    old_exist = .true.
    call map_copy_par(themap,default_map) ! Copy suggested values to Current
    error = .false.
  case default
    nuni = 10
    !
    ! back to usual code
    uvm = uvmax/(freq*f_to_k)
    !
    if (mode.eq.'TAPER') then
      if (step.eq.0.) step = sqrt(2.0)
      if (start.eq.0.) start = 10*nint(uvm/160.0)
      taper = 16*start
    else
      taper = sqrt(uv_taper(1)*uv_taper(2))
      if (taper.ne.0) then
        taper = min(taper,uvm)
      else
        taper = uvm
      endif
    endif
    if (mode.eq.'WEIGHT') nuni=18
    !
    ! Define MAP_CELL and MAP_SIZE
    taper = uvm*freq*f_to_k
    map_cell(1) = 0.02*nint(180.0*3600.0*50.0/taper/4.0)
    if (map_cell(1).le.0.02) then
      map_cell(1) = 0.002*nint(180.0*3600.0*500.0/taper/4.0)
    endif
    map_cell(2) = map_cell(1)    ! In ", rounded to 0.01" or 0.002"
    if (mode.eq.'TAPER') then
      map_size = 256
    else if (map_cell(1).gt.0.1) then
      map_size = 64
    else 
      map_size = 256
    endif
    !
    write(chain,102) uvm,uvmax*1e-3/2.0/pi
    call map_message(seve%i,rname,chain)
    write(chain,103) map_size
    call map_message(seve%i,rname,chain)
    write(chain,104) map_cell
    call map_message(seve%i,rname,chain)
    !
    ! Redefine some parameters
    map_cell = map_cell*pi/180.0/3600.0   ! In radians
    !
    ! Process sorted UV Table according to the type of beam produced
    if (allocated(drobust)) then
      deallocate(drobust,stat=ier)
      call sic_delvariable('ROBUST_BEAMS',.false.,error)
    endif
    n = 2*max(map_size(1),map_size(2))
    allocate (drobust(map_size(1),map_size(2),nuni),fft(n),stat=ier)
    if (ier.ne.0) then
      call map_message(seve%e,rname,'Memory allocation failure')
      error = .true.
      return
    endif
    if (muni.ne.nuni) then
      if (allocated(result)) deallocate(result)
      allocate(result(nuni,9))
      muni = nuni
    endif
    call uniform_beam (' ',uv_taper,   &
       &    map_size,map_cell,uniform,wcol,mcol,fft,   &
       &    error,mode,drobust,nuni,start,step,uvm,result,huv,duv)
    !
    call sic_delvariable('BEAM_SHAPE',.false.,error)
    dim(1:2) = [nuni,9]
    call sic_def_real('BEAM_SHAPE',result,2,dim,.true.,error)
    error = .false.
    !
    if (do_plot) then
      call gildas_null(hrobust)
      call gdf_copy_header(huv,hrobust,error)
      hrobust%gil%ndim = 3
      hrobust%gil%dim(1:2) = map_size
      hrobust%gil%dim(3) = nuni
      hrobust%gil%convert(:,1) = [map_size(1)/2+1d0,0d0,-dble(map_cell(1))]
      hrobust%gil%convert(:,2) = [map_size(2)/2+1d0,0d0,dble(map_cell(2))]
      hrobust%gil%convert(:,3) = 1d0
      hrobust%char%code(1) = 'ANGLE'
      hrobust%char%code(2) = 'ANGLE'
      hrobust%char%code(3) = 'ROBUST'
      hrobust%gil%xaxi = 1
      hrobust%gil%yaxi = 2
      hrobust%gil%faxi = 0
      !
      if (sic_narg(o_plot).eq.1) then
        call sic_ch(line,o_plot,1,fich,n,.true.,error)
        hrobust%file = fich
        call gdf_write_image(hrobust,drobust,error)
        !
        ! Show it
        call exec_program('@ show_robust "'//trim(hrobust%file)//'" '//mode)
      else
        call sic_mapgildas('ROBUST_BEAMS',hrobust,error,drobust)
        call exec_program('@ show_robust ROBUST_BEAMS '//mode)
      endif
    endif
    deallocate (fft)
  end select
  !
  102   format('Maximum baseline is   ',f8.1,' m,  ',f8.1,' kWavelength')
  103   format('Map size is   ',i6,' by ',i6)
  104   format('Pixel size is ',f6.3,' by ',f6.3,'"')
end subroutine uv_stat_comm
!
subroutine uv_printoffset(rname,head,data,map) 
  use clean_def
  use clean_arrays
  use gbl_message
  use imager_interfaces, only : map_message
  !-----------------------------------------------------------------  
  ! @ private
  !*
  !   IMAGER -- Support routine for UV_STAT
  !
  !     Print the list of Mosaic Offsets
  !!
  !-----------------------------------------------------------------
  character(len=*), intent(in) :: rname  !! Caller name
  type(gildas), intent(in) :: head       !! UV header
  real, intent(in) :: data(head%gil%dim(1),head%gil%dim(2))
  !! Visibilities
  type(uvmap_par), intent(in) :: map     !! Mapping parameters
  !
  ! Constants
  real(8), parameter :: pi=3.14159265358979323846d0
  real(4), parameter :: rad_to_sec = 180.0*3600.0/pi
  !
  ! Local ---
  integer :: n,j,i,valsup,valinf,lid,nval,ier
  integer, allocatable :: value(:)
  integer(kind=index_length), allocatable :: counter(:)
  character(len=60) :: chain
  !
  ! Code ----
  lid = huv%gil%column_pointer(code_uvt_id)
  if (lid.ne.0) then
    allocate(value(head%gil%dim(2)),stat=ier)
    value(:) = duv(lid,1:head%gil%dim(2)) ! Size(DUV,2) may be too large if flagged data
    valinf = minval(value)-1
    valsup = maxval(value)
    nval = valsup-valinf
    if (nval.le.1) return
    allocate(counter(nval),stat=ier)
    !
    counter = 0
    if (nval.ne.abs(map%nfields)) then
      write(*,'(A,I0,A)') 'Caution: ',nval,' pointings with NO associated coordinates:'
    else
      write(*,'(A,I0,A)') 'Multiple (',nval,') pointings:'
    endif
    do i=1,head%gil%dim(2)
      counter(value(i)-valinf) = counter(value(i)-valinf)+1
    enddo
    do i=1,nval
      write(*,'(A,I0,A,I0,A)') ' Source_ID #',i+valinf,': ',counter(i),' Visibilities'
    enddo
  endif
  !
  if (map%nfields.eq.0) return
  !
  n = abs(map%nfields)
  if (map%nfields.lt.0) then
    write(chain,'(I5,A)') n, ' Phase offsets in mosaic'
  else
    write(chain,'(I5,A)') n, ' Pointing offsets in mosaic'
  endif
  call map_message(seve%i,rname,chain)
  !
  if (allocated(hmosaic)) then
    if (map%nfields.gt.0) then
      do j=1,n-1,2
        write(*,'(2(A,F12.4,A,F12.4,A))') &
        &   '(',hmosaic(j)%opoint(1)*rad_to_sec,' ,',hmosaic(j)%opoint(2)*rad_to_sec,' )', &
        &   '(',hmosaic(j+1)%opoint(1)*rad_to_sec,' ,',hmosaic(j+1)%opoint(2)*rad_to_sec,' )'
      enddo
      if (mod(n,2).ne.0) write(*,'(A,F12.2,A,F12.2,A)') &
              &   '(',hmosaic(n)%opoint(1)*rad_to_sec,' ,',hmosaic(n)%opoint(2)*rad_to_sec,' )'
    else
      do j=1,n-1,2
        write(*,'(2(A,F12.4,A,F12.4,A))') &
        &   '(',hmosaic(j)%ophase(1)*rad_to_sec,' ,',hmosaic(j)%ophase(2)*rad_to_sec,' )', &
        &   '(',hmosaic(j+1)%ophase(1)*rad_to_sec,' ,',hmosaic(j+1)%ophase(2)*rad_to_sec,' )'
      enddo
      if (mod(n,2).ne.0) write(*,'(A,F12.2,A,F12.2,A)') &
             &   '(',hmosaic(n)%ophase(1)*rad_to_sec,' ,',hmosaic(n)%ophase(2)*rad_to_sec,' )'
    endif
  else
    do j=1,n-1,2
      write(*,'(2(A,F12.4,A,F12.4,A))') &
      &   '(',map%offxy(1,j)*rad_to_sec,' ,',map%offxy(2,j)*rad_to_sec,' )', &
      &   '(',map%offxy(1,j+1)*rad_to_sec,' ,',map%offxy(2,j+1)*rad_to_sec,' )'
    enddo
    if (mod(n,2).ne.0)     write(*,'(A,F12.2,A,F12.2,A)') &
      &   '(',map%offxy(1,n)*rad_to_sec,' ,',map%offxy(2,n)*rad_to_sec,' )'
  
  endif
end subroutine uv_printoffset
!
subroutine map_beamsize(map,huv,duv,uvmin,uvmax,plot,error)
  use image_def
  use clean_def
  use clean_default
  use gbl_message
  use gkernel_interfaces
  use imager_interfaces, except_this => map_beamsize
  !---------------------------------------------------
  ! @ private
  !*
  !  IMAGER -- Utility routine
  !
  !     Get the synthesized beam size according to
  !   current imaging parameters. Performs a quick 
  !   gridding on a representative channel.
  !!
  !---------------------------------------------------
  type(uvmap_par), intent(in) :: map    !! Mapping parameters
  type(gildas), intent(in) :: huv       !! UV Header
  real, intent(in) :: duv(:,:)          !! UV data
  real, intent(in) :: uvmin             !! Minimum UV value
  real, intent(in) :: uvmax             !! Maximum UV value
  logical, intent(in) :: plot           !! Plot beam 
  logical, intent(out) :: error         !! Logical Error flag
  !
  ! Constants
  character(len=*), parameter :: rname='MAP_BEAMSIZE'
  real(8), parameter :: pi=3.14159265358979323846d0
  real(8), parameter :: f_to_k = 2.d0*pi/299792458.d-6
  real(4) :: pixel_per_beam=4.0     ! A good compromise
  integer, parameter :: msize=256   ! Large enough for most cases
  !
  ! Local ---
  integer, save :: osize=0
  integer :: nsize, isize
  real, allocatable, save :: abeam(:,:)
  real, allocatable :: work(:) !  work(2*nsize)
  ! FFT arrays
  complex, allocatable :: ipft(:), ipf(:) !nsize*nsize), ipf(nsize*nsize)
  integer :: lx,ly,lw, nu,nv, nx,ny, ix_patch, iy_patch
  integer :: ndim, nn(2)
  real :: uvcell, map_cell, thre, rtmp
  real(kind=8) :: convert(6)
  real :: major_axis, minor_axis, pos_angle 
  !
  real, allocatable :: ipw(:), ipv(:)
  integer :: ier,jc
  integer :: grid_code = 0
  logical :: do_weig, local_error
  character(len=80) :: chain
  !
  ! Code ----
  isize = max(32,nint(uvmax/uvmin))
  call gi4_round_forfft(isize,nsize,error,map_rounding,map_power)
  nsize = min(msize,nsize)  ! Nearest acceptable number of pixels
  !
  if (nsize.eq.osize) then
    allocate(work(2*nsize),ipft(nsize*nsize),ipf(nsize*nsize),stat=ier)
  else
    if (allocated(abeam)) deallocate(abeam)
    osize = 0
    allocate(abeam(nsize,nsize),work(2*nsize),ipft(nsize*nsize),ipf(nsize*nsize),stat=ier)
  endif
  if (ier.ne.0) then
    call map_message(seve%e,rname,'Memory allocation error')
    error = .true.
    return
  endif
  osize = nsize
  !
  nu = huv%gil%dim(1)
  nv = huv%gil%nvisi
  lx = nsize  ! Grid size
  ly = nsize
  nx = nsize  ! Map size
  ny = nsize
  !
  do_weig = all(map%uniform.ne.0)
  if (do_weig) then
    allocate(ipw(nv),ipv(nv),stat=ier)
  else
    allocate(ipw(nv),stat=ier)
  endif
  if (ier.ne.0) then
    call map_message(seve%e,rname,'Memory allocation error')
    error = .true.
    return
  endif
  !
  ! That is for Natural Weighting
  jc = max(1,(2*huv%gil%nchan/3))  ! Weight channel
  lw = 7 + 3*jc                    ! Weight column 
  ipw(:) = duv(lw,:)
  !
  !
  if (do_weig) then
    write(chain,'(A,F8.2,A,F8.2)') 'Robust weighting ',map%uniform(2),' UV Cell ',map%uniform(1)
    call map_message(seve%i,rname,chain)
    ipv(:) = duv(2,:)
    local_error = .false.
    call doweig (nu,nv,   &
       &    duv,    &            ! Visibilities
       &    1,2,    &            ! U, V pointers
       &    jc,     &            ! Weight channel
       &    map%uniform(1),   &  ! Uniform UV cell size
       &    ipw,    &            ! Weight array
       &    map%uniform(2),   &  ! Fraction of weight
       &    local_error,grid_code)
    if (local_error)  then
      error = .true.
      return
    endif
  endif
  !
  do_weig = any(map%taper(1:2).ne.0)
  if (do_weig) then
    write(chain,'(A,3(F8.2))') 'Taper ',map%taper(1:3)
    call map_message(seve%i,rname,chain)
    !
    ! Should also plug the TAPER here, rather than in DOFFT later  !
    call dotape (nu,nv,   &
       &    duv,   &             ! Visibilities
       &    1,2,   &             ! U, V pointers
       &    map%taper,  &        ! Taper
       &    ipw)                 ! Weight array
  endif
  !
  ! Compute FFTs
  !
  ! Define MAP_CELL
  rtmp = uvmax*huv%gil%freq*f_to_k
  map_cell = 0.002*nint(180.0*3600.0*500.0/rtmp/pixel_per_beam)
  map_cell = map_cell*pi/180.0/3600.0
  convert = (/dble(nx/2+1),0.d0,-dble(map_cell),   &
     &    dble(ny/2+1),0.d0,dble(map_cell)/)
  !
  ! Define UV Cell size
  uvcell = 299792458D-6/huv%gil%freq/(map_cell*nx)
  !
  ! Quick, simple minded gridding
  call doqfft (nu,nv,   &      ! Size of visibility array
     &    duv,          &      ! Visibilities
     &    1,2,lw,       &      ! U, V pointers
     &    lx,ly,        &      ! Cube size
     &    ipft,         &      ! FFT cube
     &    ipw,          &      ! Weight array
     &    uvcell)              ! UV cell size
  !
  ! Make beam, not normalized
  ndim = 2
  nn(1) = nx
  nn(2) = ny
  call extracs(1,nx,ny,1,ipft,ipf,lx,ly)
  call fourt  (ipf,nn,ndim,-1,1,work)
  call cmtore (ipf,abeam,nx,ny)
  !
  ! Normalize beam - no grid correction needed here
  rtmp = abeam(nx/2+1,ny/2+1)
  abeam = abeam/rtmp
  !
  ! Get Fitting Threshold 
  rtmp  = minval(abeam(:,:)) 
  thre = max(min(abs(-1.5*rtmp),0.7),0.3)
  !
  ! Fit beam
  ix_patch = nx/2
  iy_patch = ny/2
  major_axis = 0.
  minor_axis = 0.
  pos_angle  = 0.
  call fibeam (rname,abeam,nx,ny,   &
   &      ix_patch,iy_patch,thre,   &
   &      major_axis,minor_axis,pos_angle,   &
   &      convert,error)
  !
  if (plot) then
    call gr_exec('CLEAR')
    call gr4_rgive(nx,ny,convert,abeam)
    call gr_exec('LIMITS /RG')
    call gr_exec('SET BOX SQUARE')
    call gr_exec('PLOT')
    call gr_exec('BOX /UNIT SEC')
    call gr_exec('LABEL "Offset (`)" /X')
    call gr_exec('LABEL "Offset (`)" /Y')
    call gr_exec('WEDGE')
    write(chain,*) 'LET TRUE_SCALE ',rtmp,1.0
    call exec_command(chain,local_error)
  endif
end subroutine map_beamsize
!
subroutine uv_stat_base(comm,huv,duv,bmin,bmax,jcol,plot,error)
  use image_def
  use gkernel_interfaces
  use imager_interfaces, only : map_message, my_histo44
  ! @ private
  !*
  ! IMAGER -- Support routine for  
  !     UV_STAT BASELINE /PLOT
  !!
  character(len=*), intent(in) :: comm  !! Calling command
  type(gildas), intent(in) :: huv       !! UV header
  real, intent(in) :: duv(:,:)          !! Visibilities
  real, intent(in) :: bmin              !! Shortest baseline
  real, intent(in) :: bmax              !! Longest baseline
  integer, intent(in) :: jcol           !! Weight column
  logical, intent(in) :: plot           !! Display weight distribution ?
  logical, intent(inout) :: error       !! Logical error flag
  !
  ! Constants
  real(8), parameter :: clight=299729458d0
  real(8), parameter :: pi=acos(-1d0)
  real(8), parameter :: rad_to_sec=180*3600/pi
  !
  ! Local ---
  integer :: nhist,iv,ih,i10,i85
  real, allocatable :: histo(:,:)
  integer :: nvisi
  integer :: ier
  real :: hmin,hmax,bstep,v,b,mrs,hrs,sumw,cumw
  real(8) :: umin,umax,lambda
  character(len=80) :: chain
  !
  ! Code ----
  nvisi = huv%gil%nvisi
  nhist = nint(bmax/bmin*4.0)
  !
  allocate(histo(nhist,2),stat=ier)
  !
  bstep=(bmax-bmin)/(nhist-1)
  if (bstep.le.0) return ! Avoid big issues
  !
  hmin=bmin-bstep/2.0
  hmax=bmax+bstep/2.0
  histo = 0.
  do iv=1,nvisi
    b = sqrt(duv(1,iv)**2+duv(2,iv)**2)
    if (b.ge.hmin .and. b.le.hmax) then
      v = (b-bmin)/bstep
      ih = int(v)+1
      if (jcol.eq.0) then
        histo(ih,2) = histo(ih,2)+1
      else
        histo(ih,2) = histo(ih,2)+max(0.,duv(jcol,iv))
      endif
    endif
  enddo
  do ih=1,nhist
    histo(ih,1) = bmin+ih*bstep
  enddo
  !
  call gr_exec('CLEAR')
  call gr4_give('X',nhist,histo(:,1))
  call gr4_give('Y',nhist,histo(:,2))
  call gr_exec('LIMITS')
  call gr_exec('SET BOX SQUARE')
  call gr_exec('HISTO')
  call gr_exec('AXIS XL')
  call gr_exec('AXIS YL')
  call gr_exec('AXIS YR')
  call sic_get_dble('USER_XMIN',umin,error)
  call sic_get_dble('USER_XMAX',umax,error)
  lambda = clight/huv%gil%freq*1D-6
  umin = umin/lambda*1D-3
  umax = umax/lambda*1D-3
  write(chain,'(A,1PG11.4,X,1PG11.4,A)') 'AXIS XU ',umin,umax,' /LABEL P'
  call gr_exec(chain)
  call gr_exec('DRAW TEXT 0 2 "Baseline (k\gl)" 5 0 /CHAR 8')
  call gr_exec('LABEL "Baseline (m)" /X')
  call gr_exec('LABEL "Weights" /Y')
  !
  ! Make a Gaussian fit
  b = (bmax+bmin)/2.0
  v = maxval(histo(:,2))
  write(chain,'(A,1PG11.4,1X,1PG11.4)') 'mfit y=&A*gaussian1d(X,0,&B) /start',v,b 
  call exec_command(chain,error)
  call gr_exec('PENCIL 1')
  call gr_exec('CURVE X MFIT%FIT')
  call gr_exec('PENCIL 0')
  call sic_get_real('MFIT%PAR[2]',v,error)
  b = 1.2*lambda/v  
  b = b*rad_to_sec
  write(chain,'(A,F8.3,A)') 'DRAW TEXT 0 -2 "Estimated beam size ',b,' `" 5 0 /CHAR 8'
  error = .false.
  !  call exec_command(chain,error) ! Bug - it crashes because of attempts to re-write the command
  call gr_exec1(chain)  ! Equivalent, but working
  !
  ! Compute the Weight Percentage
  sumw = sum(histo(:,2))
  cumw = 0.
  histo(:,2) = histo(:,2)/sumw
  do ih=1,nhist
    cumw = cumw+histo(ih,2)
    if (cumw.le.0.1)  then
      i10 = ih
    else if (cumw.le.0.85) then
      i85 = ih
    endif
  end do
  !Print *,'I10 ',i10,histo(i10,1)
  !Print *,'I85 ',i85,histo(i85,1)
  !
  mrs = lambda/histo(i10,1)*rad_to_sec 
  hrs = lambda/histo(i85,1)*rad_to_sec
  write(chain,'(A,F8.3,A)') 'DRAW TEXT 0 -3 "Recoverable scale    ',mrs,' `" 5 0 /CHAR 8'
  call gr_exec1(chain)  ! Equivalent, but working
  write(chain,'(A,F8.3,A)') 'DRAW TEXT 0 -4 "Estimated Resolution ',hrs,' `" 5 0 /CHAR 8'
  call gr_exec1(chain)  ! Equivalent, but working
end subroutine uv_stat_base
!
